// main.h
// global include file for open-source proportional solenoid valve ventilator
//
// Copyright 2020 Bill Dally
// 
// Licensed under the Ventilator Open Source License (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
// ./LICENSE
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//


#include <avr/io.h>
#include <util/delay.h>
#include "params.h"

// Version
#define VERSION "3.5 9/20/20"

#define BLEND 1 // 1 if O2 blending 0 for single gas

// default parameter values
#define DEF_BPM 15
#define DEF_VOL 400
#define DEF_ETOI 20
#define DEF_MAXP 400 // mm H2O
#define DEF_TRIGP 25 // mm H2O - negative offset from PEEP
#define TRIG_TIME 30 // 30ms
#define DEF_SPEEP 0 // default set peep
#define DEF_FIO2 21 // default fraction inspired oxygen

// valve specific PWM parameters
#define MIN_PWM 600          // point at which flow starts
#define MAX_PWM 1200         // 
#define MAX_FLOW_MLPS 1167   // in mL/sec
#define MAX_FLOW_MLPM 70000  // in mL/min
#define SENSOR_MAX_FLOW 1667 // ml/sec - for 100 L/m sensors

#define TICK 1 // ms per RTC invocation

#define MAX_LOG 256 // for logging
#define MAX_LOG_CHAN 11 //

// Control
// 48 4 8 / 16 worked on one unit
// 0 1 0 / 16 (or 0 2 0 / 32) is super reliable
#define PGAIN 16  // 
#define IGAIN 2 // was 4 - 4 is good for flows 150 and above
#define DGAIN 16  // 
#define PRESS_GAIN 200 
#define PRESS_DGAIN 200 
#define PRESS_IGAIN 2 
#define GAIN_DIV_LG2 6 // gains are in 1/64s
#define GAIN_DIVISOR (1<<GAIN_DIV_LG2)

#define PDOWN_GAIN 4 // extra effort to avoid overshoot
#define PDOWN_GAIN2 15 // extra effort to avoid overshoot - at second threshold
#define PDOWN_GAIN3 100 // extra effort to avoid overshoot - at second threshold

#define PDOWN_MARGIN -3 // point where extra effort kicks in
#define PDOWN_MARG2 -6 // point where even more kicks in
#define PDOWN_MARG3 -10 // point where enormous gain kicks in

#define PDOWN_CUTOFF 15 // point above maxp where cflow is cut off

// guard band of 32mm H2O to switch from flow to pressure regulation
#define PGUARD_LG2 5                      //
#define PGUARD (1<<PGUARD_LG2)            // guard band on pressure regulation - must be a power of 2
#define HALF_PGUARD (1<<(PGUARD_LG2-1))   // 

// Times
#define TIME_FF 70 // time for feed forward before engaging PID
#define TGAP 200 // time from end of ton to doing end of breath activities
#define TOFFSET 60 // turn on delay - turn off delay - in ms
#define TMAX_MUL 1.15 // maximum extension to make volume - 1.15 allows time for rise time and 10% variation
#define INSPIRATORY_PAUSE_TIME 130 // ms - ton shortened by this amount to allow time for diaphragm valve operation
#define TRISE 20 // ms
#define TPULSE 3 // ms - time to pulse valve at full PWM to break stickiness


// to smooth pressure with IIR filter
// smooth = (new + smooth * (2^n-1)) * 2^-n
#define PS_SHIFT 3
#define PS_CONST ((1<<PS_SHIFT)-1)

// to limit windup
#define MAX_IERR 8000
#define MIN_IERR -8000
#define MAX_PIERR 12000
#define MIN_PIERR -12000

// I/O Definitions
// LEDs are on Port C
#define LED1_PC 5 
#define LED2_PC 6 
#define LED3_PC 7
//#define ALARM_PC 4 - stole alarm bit for mux control
#define FLOW_MUX1_PC 4

// PWM on Port D
#define OC1A_PD 5
#define OC1B_PD 4

// Solenoids 3 and 4 on Port D
#define SOL3_PD 6
#define SOL4_PD 7

// Encoders on Port B
// These reversed from 2.0 to 2.1
#define ENC11_B 1
#define ENC12_B 0
#define ENC21_B 3
#define ENC22_B 2

//#define SW1_B 4 // switch is  PD2 - stole this for mux
#define FLOW_MUX0_PB 4
#define SW2_C 2 // encoder switches
#define SW3_C 3

// EE PROM Locations
// zero calibration
#define EER_SIARGO_ZERO 2 // 1 word
#define EER_NXP_ZERO 4 // 1 word
#define EER_HP_ZERO 6 // honeywell pressure
#define EER_HP2_ZERO 8 // honeywell pressure 2

// saved parameters - 22 Bytes + alarm
#define EER_RUN 20 // run - stored redundantly
#define EER_MODE 21 // mode
#define EER_VOL 22 // tidal volume
#define EER_BPM 24 // RR
#define EER_ETOI 25 // I:E
#define EER_MAXP 26 // maxp
#define EER_SPEEP 28 // set peep
#define EER_PSF 30 // flow fraction to end pressure support
#define EER_TRIGP 32 // trigger pressure
#define EER_ALARMP 34 // trigger pressure
#define EER_LMV 36 // low minute volume
#define EER_HMV 38 // high minute volume
#define EER_DISCON_P 40 // disconnect pressure
#define EER_ALARM 42 // current state of alarms

// offsets for saved patients - add to numbers above to get locations
#define EER_OFFSET_0 (44-EER_RUN) // 
#define EER_OFFSET_1 (66-EER_RUN) // 
#define EER_OFFSET_2 (88-EER_RUN) // 
#define EER_OFFSET_3 (110-EER_RUN) // 

#define MAX_EER_OFFSET 3  // 0 to 3 to select

// flow to pwm table
#define EER_F2P 2048 // F2P_SZ * 2 locations

#define EER_O2_F2P 4096 // F2P_SZ * 2 locations

// Modes
#define MODE_VOL 0
#define MODE_PRESS 1
#define MODE_VOL_ASSIST 2
#define MODE_PRESS_ASSIST 3
volatile uint8_t mode ;         // operating mode - volume or pressure control

// volume expiratory states
#define VE_START 0
#define VE_REG 1
//#define PEEP_ON_MARGIN -30
#define PEEP_HOLD_MARGIN 10
#define CAP_FLOW 200
#define PEEP_CLOSE_FRAC 0.5 // initial value - gets adjusted
#define PEEP_HIGH_MARGIN 1.5

// Pressure control
// states
#define PS_RISING 0
#define PS_RISE2 1
#define PS_RISE3 2
#define PS_RISE4 3
#define PS_HOLD 4
#define PS_IDLE 5

#define PS_FRAC 2  // numerator of off fraction
#define PS_LG_DENOM 3 // denom is 8

int16_t ps_frac ; // fraction of flow at which to stop pressure support

#define PS_FLOW 1100 // flow for rising edge - was 800
#define PS_FLOW2 500 // flow to ramp down to
#define PS_FLOW_DEC 8 // amount to decrement per ms

#define PS_DELAY 30 // ms minimum regulation until hold
#define PS_DELAY2 30 // ms to sample flow after hold

#define PS_GUARDBAND 20
#define PS_GB2 5
#define PS_FUZZ 5

// was 1000 - gave a big glitch - gets divided by 16
// 5000 gives too much overshoot - which doesn't come down
// 1000 gives big glitch
// 3000 gives a small overshoot
// This really should be related to the flow needed to hold pressure on last breath
#define PS_INIT_IERR 3000

#define ABS_PRESS_LIMIT 600 // mm H2O - point at which exhale valve opens
#define NEG_PRESS_LIMIT -60 // mm H2O - point at which exhale valve opens

// structure for PID controller -
struct pid {
  int16_t err ; // error - updated every time step
  int16_t ierr ; // integral error
  int16_t derr ; // differnetial error
  int16_t oerr ; // old error - for computing derr
  int16_t o2err ; // two time steps back
  int16_t cflow ; // commanded flow
} ;

// separate PIDs for air and o2
struct pid air_pid ;
struct pid o2_pid ;
struct pid press_pid ;

volatile uint8_t pstate ;       // pressure control state
volatile uint8_t vestate ;      // volume expiratory period state
volatile int16_t ps_flow ;      // peak flow
volatile int16_t ps_off_flow ;  // flow to turn off at
volatile int16_t ps_timer ;     // timer for flow sampling

volatile uint8_t sol3_state ;   // state of sol3
volatile uint8_t sol4_state ;   // state of sol4

// derived values
volatile uint16_t tcy ;         // cycle time ms
volatile uint16_t ton ;         // inhale time ms
volatile uint16_t toff ;        // time when volume completed
volatile uint16_t tarm ;        // point at which patient can initiate
volatile uint16_t tdown ;       // time to start ramp down
volatile uint16_t trise ;      // rise time
volatile uint16_t tmax ;        // maximum extension to make TV
volatile int16_t flow ;        // flow during inhale in mlps 0-833 at 50L/min
volatile uint16_t flow_mlpm ;   // flow during inhale in mlpm 0-50000 at 50L/min
volatile int16_t flow_step ;   // step of flow during rise - per TICK
volatile int16_t tflow ;   // target flow
volatile int16_t to_flow ;   // target o2 flow
volatile int16_t ta_flow ;   // target air flow
volatile int32_t vol_limit ;   // tidal volume limit - multiplied by integration factor 1000/TICK

// control values
volatile int16_t cflow ;       // commanded flow
volatile int32_t lcflow ;      // 32b version for calculations
volatile int16_t mflow ;       // measured flow
volatile int16_t a_flow ;       // target air flow
volatile int16_t o_flow ;       // target o2 flow
volatile int16_t ca_flow ;       // commanded air flow
volatile int16_t co_flow ;       // commanded o2 flow
volatile int16_t ma_flow ;       // measured air flow
volatile int16_t mo_flow ;       // measured o2 flow
volatile int16_t error ;       // error - signed value
volatile int16_t d_err ;       // derivative of error - delta since last TICK
volatile int16_t i_err ;       // integral of error
volatile int16_t oerror ;      // one back for derivative
volatile int16_t o2error ;     // two back to smooth derivative

// pressures
volatile int16_t mpressure ;   // measured pressure mm H2O - signed
volatile int16_t smpressure ;  // smoothed measured pressure - avoids ringing on transitoins - can be 200mm
volatile int16_t osmpressure ;  // smoothed measured pressure - avoids ringing on transitoins - can be 200mm
volatile int16_t oosmpressure ;  // smoothed measured pressure - avoids ringing on transitoins - can be 200mm
volatile int16_t peep ;        // peep pressure - recorded before start of breath
volatile int16_t set_peep ;    // target peep pressure
volatile int16_t set_peep_close ; // fraction of target peep where expiration valve closes
volatile int16_t set_peep_open ; // fraction of target peep where expiration valve opens
volatile int16_t pip ;         // maximum inspiratory pressure this breath - running maximum
volatile int16_t bpip ;        // maximum inspiratory pressure for last breath - set at end of breath
volatile int16_t press_err ;   // pressure error
volatile int16_t opress_err ;  // old pressure error
volatile int16_t press_derr ;  // diff pressure error
volatile int16_t ferr ;        // attenuated error in guard-band 
volatile int16_t frac ;        // fraction attenuated
volatile int16_t fracx ;       // 1-frac
volatile int16_t cmode ;       // how gain is calculated

// for PEEP setting
#define SPC_INC 2
#define PEEP_FUZZ 5
#define PEEP_BIG_ERR 50
volatile uint8_t spc_high ;     // set_peep_close was high on this breath


// pwm visible for table refinement
volatile uint16_t pwm ; // could also just read OCR3A

// control parameters
// gains are in 1/16ths - shifted by 4 
volatile int16_t pgain ; // proportional gain
volatile int16_t dgain ; // derivative
volatile int16_t igain ; // integral
volatile int16_t press_gain ; // pressure proportional gain
volatile int16_t press_dgain ; // pressure diff gain
volatile int16_t press_igain ; // pressure diff gain

// timing
#define LG_GTICK 8 
#define GTICK (1<<LG_GTICK)
#define GTICK_MASK (GTICK - 1)
#define GTICKS_PER_MIN 234  // (60,000/GTICK) - manually recalculate
volatile uint16_t time ;    // time in current cycle - in ms
volatile uint16_t gtime ;   // global time - in GTICK increments -

// measured volume
// add mL/s each TICK ms -
// 1000/TICK times the volume
// divide by TICK/1000
#define MVOL_VOLx256k  262     // TICK/1000 * 2^18 - manually recalculate if you change TICK
volatile uint32_t mvol ;       // integrated flow over breath - mL/s each TICK ms
volatile uint32_t ramp_vol ;   // volume expect
volatile uint16_t bvol ;       //  volume at end of breath - divide by 1000
volatile uint16_t minute_vol ; // volume for last minute
volatile uint16_t emvol ;      // expected minute volume

// to compute minute volume - history of last N breaths
// one minute history
// volume and pip for now - can add more channels later
#define MAX_BHIST 3 
#define BHIST_BVOL 0
#define BHIST_BPIP 1
#define BHIST_GTIME 2
volatile int16_t b_hist[MAX_BPM][MAX_BHIST] ;
volatile uint8_t b_hist_ptr ; // points to last valid location
volatile uint8_t b_hist_n_val ; // number of valid entries - only relevant during startup


// log
volatile uint8_t log_on ;
volatile uint16_t log_ptr ;
volatile uint8_t log_step ; // ticks per log element
volatile uint16_t ltime ; // ticks since last log
volatile int16_t flog[MAX_LOG][MAX_LOG_CHAN] ;

// pressure and flow log - every 32ms - ~32 per second
#define MAX_PF_LOG 1200
#define PF_LOG_MASK 0x1F 
#define PF_LOG_STEP 32
#define PF_LOG_PEEP_OFFSET 4 // reach back 4 entries to get PEEP

volatile int16_t pf_log[MAX_PF_LOG][3] ;
volatile int16_t pf_log_ptr ;
volatile int16_t pf_log_cnt ;

volatile uint8_t  run  ;     // run ventilator at current settings if true
volatile uint8_t  ftest ;    // test signal to RTC
volatile uint8_t  trun ;     // test signal to RTC


// input values
uint8_t bpm  ;     // breaths per minute
uint16_t vol  ;   // tidal volume in ml
uint8_t etoix10  ; // e:i ratio times 10
uint16_t maxp ;  // max pressure - mm H2O
uint8_t fio2 ;   // 21-100% - inspired O2

// for patient trigger
int16_t trigp ;     // trigger pressure - negative offset from PEEP
uint8_t p_trig_on ; // enable patient trigger
uint16_t trig_time ; // patient trigger counter

int16_t in_p ; // input pressure
int16_t o2_p ; // O2 pressure

int16_t in_v ; // input voltage

// mixing - these are x1024
uint16_t frac_o2_x1k ; // fraction of o2 needed to achieve set fio2 - 0 for 21% fio2
uint16_t frac_air_x1k ; // fraction of air needed to achieve set fio2 - 1 for 21%

// alarm thresholds
#define DISCONNECT_COUNT_THRESHOLD 3  // constant for now
#define PIP_DISCONNECT 50             // mm H2O
#define PIP_ALARM_MARGIN 50

#define ALARM_MIN_IN_P 100            // 10PSI
#define ALARM_APNEA_COUNT 2
#define ALARM_MIN_IN_V 11

volatile uint16_t low_minute_vol ; 
volatile uint16_t high_minute_vol ; 
volatile uint16_t pip_disconnect ;
volatile int16_t alarm_pip ;
volatile int16_t trig_pip ;

// alarm codes
#define ALARM_MAXP            1
#define ALARM_DISCONNECT      2
#define ALARM_LOW_MINUTE_VOL  4
#define ALARM_HIGH_MINUTE_VOL 8
#define ALARM_APNEA          0x10
#define ALARM_IN_P           0x20
#define ALARM_IN_V           0x40
#define ALARM_FLOW           0x80
#define ALARM_RESET          0x100
#define ALARM_TEST           0x200

// apnea count
volatile uint8_t apnea_count ;

// disconnect count
volatile uint8_t disconnect_count ;

// alarm variable - or of all active codes
volatile uint16_t alarm ;

// exactly once variable for callback
uint8_t once ;

uint8_t front_panel_lock ;

// slow versions of variables to keep LCD from blurring
int16_t xsmpress ;
uint16_t xtime ;
int16_t xmflow ;

// output pressure and flow every 32ms if true
int8_t serial_pf ;

// export
void set_leds(uint8_t val) ;


