// restore.c
// Routines to save and restore settings to EE PROM
//
// (c) Bill Dally, May 2020
//
// Copyright 2020 Bill Dally
// 
// Licensed under the Ventilator Open Source License (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
// ./LICENSE
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <avr/io.h>
#include "main.h"
#include "two_wire.h"
#include "serial.h"
#include "ui.h"

// compute offset for buffer number
uint16_t restore_int_to_offset(uint8_t n) {
  uint16_t off ;
  switch(n) {
  case 0:
    off = EER_OFFSET_0 ;
    break ;
  case 1:
    off = EER_OFFSET_1 ;
    break ;
  case 2:
    off = EER_OFFSET_2 ;
    break ;
  case 3:
  default:
    off = EER_OFFSET_3 ;
    break ;
  }
  return(off) ;
}

// save run redundantly
void save_run(uint8_t run) {
  uint8_t srun ;
  if(run ==1) srun = 0xA5 ;
  else srun = 0x5A ;
  //put_string_P(PSTR("R = ")) ; put_byte(srun) ; crlf() ; // for test
  write_at_byte_sa(EER_RUN, srun) ;
}

// save parameters at offset
void save_parameters(uint16_t offset) {
  write_at_byte_sa(EER_MODE + offset, mode) ;
  write_at_word_sa(EER_VOL + offset, vol) ;
  write_at_byte_sa(EER_BPM + offset, bpm) ;
  write_at_byte_sa(EER_ETOI + offset, etoix10) ;
  write_at_word_sa(EER_MAXP + offset, maxp) ;
  write_at_word_sa(EER_SPEEP + offset, set_peep) ;
  write_at_word_sa(EER_PSF + offset, ps_frac) ;
  write_at_word_sa(EER_TRIGP + offset, trigp) ;
  write_at_word_sa(EER_ALARMP + offset, alarm_pip) ;
  write_at_word_sa(EER_LMV + offset, low_minute_vol) ;
  write_at_word_sa(EER_HMV + offset, high_minute_vol) ;
  write_at_word_sa(EER_DISCON_P + offset, pip_disconnect) ;
}

void restore_parameters(uint16_t offset) {
  mode =  read_at_byte_sa(EER_MODE + offset) ;
  vol = read_at_word_sa(EER_VOL + offset) ;
  bpm = read_at_byte_sa(EER_BPM + offset) ;
  etoix10 = read_at_byte_sa(EER_ETOI + offset) ;
  maxp = read_at_word_sa(EER_MAXP + offset) ;
  set_peep = read_at_word_sa(EER_SPEEP + offset) ;
  ps_frac = read_at_word_sa(EER_PSF + offset) ;
  trigp = read_at_word_sa(EER_TRIGP + offset) ;
  alarm_pip = read_at_word_sa(EER_ALARMP + offset) ;
  low_minute_vol = read_at_word_sa(EER_LMV + offset) ;
  high_minute_vol = read_at_word_sa(EER_HMV + offset) ;
  pip_disconnect = read_at_word_sa(EER_DISCON_P + offset) ;
  compute_derived() ;
}

// on starting vent save all settings - including run
void save_settings(void) {
  put_string_P(PSTR("Save S\n\r")) ;
  save_parameters(0) ; // no offset for main set
  write_at_word_sa(EER_ALARM, alarm) ; // save current alarms
}

// on stopping the ventilator record that run is zero
void save_stop(void) {
  save_run(0) ;
}

// on reset, check saved state of run, if its 1 restore all settings - and set alarm
void restore_settings(void) {
  uint8_t srun = read_at_byte(EER_RUN) ;
  if(srun == 0xA5) { // if running when reset
    restore_parameters(0) ; // restore parameters first
    alarm = read_at_word_sa(EER_ALARM) ; // and alarm state
    run = 1 ; // then restart
    alarm |= ALARM_RESET ; // and set reset alarm
  }
}
