// ui.c
// user interface for open-source proportional solenoid valve ventilator
// (c) Bill Dally, April 2020
//
// Copyright 2020 Bill Dally
// 
// Licensed under the Ventilator Open Source License (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
// ./LICENSE
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <stdio.h>
#include "main.h"
#include "params.h"
#include "serial.h"
#include "flow.h"
#include "siargo.h"
#include "pwm.h"
#include "nxp_pressure.h"
#include "honeywell_pressure.h"
#include "adc.h"
#include "lcd.h"
#include "control.h"
#include "ui.h"
#include "alarm.h"
#include "restore.h"

// for encoder interface
#define NR_ENC1  26
#define RUN_ENC1 22

// 8 per line
// These are things that can be selected to observe and/or change with Encoder 1
char *enc1_vals[NR_ENC1] = {"Mode","TV", "RR", "I:E", "MaxP", "FIO2", "Flow",
			    "Set PEEP","% Flow",
			    "Spont P",
			    "Alarm P","L Min V", "H Min V",
			    "Dis P", 
			    "Last TV", "Minute Vol", "PIP", "PEEP",
			    "Air P","O2 P","In V", "Version",
			    "Save", "Restore",
			    "Alarm Test","Cal"} ;

// These must line up with the strings above
#define ENC1_MODE    0
#define ENC1_TV      1
#define ENC1_RR      2
#define ENC1_IE      3
#define ENC1_MAXP    4
#define ENC1_FIO2    5
#define ENC1_FLOW    6 
#define ENC1_SPEEP   7 
#define ENC1_PCTF    8 
#define ENC1_SPONTP  9 
#define ENC1_ALARMP  10 
#define ENC1_LMINV   11 
#define ENC1_HMINV   12 
#define ENC1_DISCONP 13 
#define ENC1_LTV     14 
#define ENC1_MVOL    15 
#define ENC1_PIP     16 
#define ENC1_PEEP    17 
#define ENC1_INP     18 
#define ENC1_O2P     19 
#define ENC1_INV     20 
#define ENC1_VERS    21 
#define ENC1_SAVE    22 
#define ENC1_RESTORE 23 
#define ENC1_ALARM_T 24 
#define ENC1_CAL     25

#define CAL_CCODE_ZERO 9
#define CAL_CCODE_TAB 11



int8_t enc1_ptr = 0; // signed to handle wrap correctly
uint8_t enc1 = 0 ;
uint8_t oenc1 = 0 ;
uint8_t enc2 = 0 ;
uint8_t oenc2 = 0 ;
uint8_t sw1 = 0 ;
uint8_t osw1 = 0 ;
uint8_t sw2 = 0 ;
uint8_t osw2 = 0 ;
uint8_t sw3 = 0 ;
uint8_t osw3 = 0 ;
int16_t enc1_count = 0 ; // these are updated by the RTC callback
int16_t enc2_count = 0 ;
int16_t enc1_a_count = 0 ; // these are acknowledged by foreground
int16_t enc2_a_count = 0 ;

uint8_t eer_select = 0 ;
uint8_t ccode = 0 ;


// put 1/100 of a value 
void l_put_100th(uint16_t time) {
  l_put_word_x(time/100, 1, 100) ;
}

// Row 4 - what used to be on Row 3
void lcd_row4(void) {
  //lcd_goto(ROW3) ;
  lcd_buff_init() ; // clear buffer
  if(alarm == 0) {
    if(run == 0) { // display derived parameters
      l_put_char('F') ;
      l_put_word(flow) ;
      l_put_string_P(PSTR(" t")) ;
      l_put_100th(tcy) ; // put time in tenths
      l_put_char(' ') ;
      l_put_100th(ton) ;
      l_put_string_P(PSTR(" M")) ;
      l_put_word(emvol/10) ; // expected minute volume 
    } else { // run = 1
      if(mode == MODE_BLEND) {
	l_put_string_P(PSTR("F ")) ;
	l_put_word(bflow) ;
	l_put_string_P(PSTR("A ")) ;
	l_put_word(ma_flow) ;
	l_put_string_P(PSTR("O2 ")) ;
	l_put_word(mo_flow) ;
      } else {
	l_put_string_P(PSTR("TV")) ;
	l_put_word(bvol) ;
	l_put_string_P(PSTR(" P")) ;
	l_put_word(bpip/10) ;
	l_put_string_P(PSTR(" E")) ;
	l_put_word(peep/10) ;
	l_put_string_P(PSTR(" M")) ;
	l_put_word(minute_vol/10) ;
      }
    }
  } else { // alarm display
    if(alarm & ALARM_MAXP) {
      l_put_string_P(PSTR("PRESSURE ")) ;
      l_put_word(pip) ;
      //l_put_char(' ') ;
    }
    else if(alarm & ALARM_DISCONNECT) l_put_string_P(PSTR("DISCONNECT ALARM")) ;
    else if(alarm & ALARM_APNEA) l_put_string_P(PSTR("APNEA ALARM")) ;
    else if(alarm & ALARM_LOW_MINUTE_VOL) {
      l_put_string_P(PSTR("LOW MIN VOL ")) ;
      l_put_word(minute_vol) ;
      //l_put_char(' ') ;
    } else if(alarm & ALARM_HIGH_MINUTE_VOL) {
      l_put_string_P(PSTR("HIGH MIN VOL ")) ;
      l_put_word(minute_vol) ;
      //l_put_char(' ') ;
    } else if(alarm & ALARM_FLOW) {
      l_put_string_P(PSTR("FLOW SENSOR ERROR")) ;
    } else if(alarm & ALARM_IN_P) {
      l_put_string_P(PSTR("LOW IN PRESS ")) ;
      l_put_word(in_p/10) ;
      l_put_char('.') ;
      l_put_word(in_p % 10) ;
      //l_put_char(' ') ;
    } else if(alarm & ALARM_RESET) l_put_string_P(PSTR("RESET")) ;
    else if (alarm & ALARM_TEST) l_put_string_P(PSTR("TEST")) ;
  }
  lcd_buff_row(ROW4) ;
}
  
void put_l4(const char *s) {
  lcd_buff_init() ;
  l_put_string_P(s) ;
  lcd_buff_row(ROW4) ;
}


// Old Row 4 - no longer used - maybe fit some on Row 2
/*
void lcd_row4(void) {
  //lcd_goto(ROW4) ;
  lcd_buff_init() ;
  if(alarm == 0) {
    if(run == 1) {
      l_put_char('R') ;
      if(time < ton) l_put_char('I') ;
      else l_put_char('E') ;
      l_put_space() ;
      l_put_char('t') ;
      l_put_word_x(xtime/100, 1, 100) ; // time in tenths
      l_put_space() ;
      l_put_char('P') ;
      l_put_word_x(xsmpress/10,1,10) ; // pressure in cm
      l_put_space() ;
      l_put_char('F') ;
      l_put_word_x(xmflow/10,1,10) ; // flow
    }
    else l_put_char('S') ;
  } else { // display highest priority alarm
    if(alarm & ALARM_MAXP) {
      l_put_string_P(PSTR("PRESSURE ")) ;
      l_put_word(pip) ;
      //l_put_char(' ') ;
    }
    else if(alarm & ALARM_DISCONNECT) l_put_string_P(PSTR("DISCONNECT ALARM")) ;
    else if(alarm & ALARM_APNEA) l_put_string_P(PSTR("APNEA ALARM")) ;
    else if(alarm & ALARM_LOW_MINUTE_VOL) {
      l_put_string_P(PSTR("LOW MIN VOL ")) ;
      l_put_word(minute_vol) ;
      //l_put_char(' ') ;
    } else if(alarm & ALARM_HIGH_MINUTE_VOL) {
      l_put_string_P(PSTR("HIGH MIN VOL ")) ;
      l_put_word(minute_vol) ;
      //l_put_char(' ') ;
    } else if(alarm & ALARM_IN_P) {
      l_put_string_P(PSTR("LOW IN PRESS ")) ;
      l_put_word(in_p/10) ;
      l_put_char('.') ;
      l_put_word(in_p % 10) ;
      //l_put_char(' ') ;
    } else if(alarm & ALARM_RESET) l_put_string_P(PSTR("RESET")) ;
    else if (alarm & ALARM_TEST) l_put_string_P(PSTR("TEST")) ;
  }
  lcd_buff_row(ROW4) ;
}
*/

void l_put_etoi(void) {
  l_put_word(etoix10/10) ;
  l_put_char('.') ;
  l_put_word(etoix10 %10) ;
}

/*
void lcd_put_etoi(void) {
  lcd_put_word(etoix10/10) ;
  lcd_put_char('.') ;
  lcd_put_word(etoix10 %10) ;
}
*/
  
void lcd_params(void) {
  lcd_buff_init() ;
  char c = 'V' ;
  //lcd_home() ;
  //  lcd_clear() ;
  switch(mode) {
  case MODE_VOL:
    c = 'V' ;
    break ;
  case MODE_VOL_ASSIST:
    c = 'v' ;
    break ;
  case MODE_PRESS:
    c = 'P' ;
    break;
  case MODE_PRESS_ASSIST:
    c = 'p' ;
    break ;
  case MODE_BLEND:
    c = 'C' ; // for constant flow
    break ;
  }
  l_put_char(c) ;
  l_put_space() ;
  if(mode == MODE_BLEND) {
    l_put_char('F') ;
    l_put_word(bflow) ;
  } else {
    l_put_char('V') ;
    l_put_word(vol) ;
  }
  l_put_space() ;
  l_put_char('R') ;
  l_put_word_x(bpm,1,10) ;
  l_put_string_P(PSTR(" EI")) ;
  l_put_etoi() ;
  l_put_space() ;
  l_put_char('P') ;
  l_put_word(maxp/10) ;
  lcd_buff_row(ROW1) ;

  // and put PEEP, FIO2, and flow on row 2
  lcd_buff_init() ;
  if(BLEND) {
    l_put_char('O') ;
    l_put_word((uint16_t) fio2) ;
  }
  l_put_space() ;
  l_put_char('E') ;
  l_put_word(set_peep/10) ;
  l_put_space() ;
  l_put_char('F') ;
  l_put_word((100*ps_frac) >> PS_LG_DENOM) ;
  lcd_buff_row(ROW2) ;
}
 
int8_t enc_to_inc(uint8_t enc, uint8_t oenc) { // 0, 1, 3, 2
  if(enc == oenc) return(0) ;
  switch(enc) {
  case 0:
    if(oenc == 2) return(1) ;
    break ;
  case 1:
    if(oenc == 0) return(1) ;
    break ;
  case 2:
    if(oenc == 3) return(1)  ;
    break ;
  case 3:
    if(oenc == 1) return(1) ;
    break ;
  }
  return(-1) ;
}

// read encoders - if changed from last value, update counts
// this is called from the RTC callback
void update_encoders(void) {
  uint8_t pinb ;
  uint8_t pinc ;
  uint8_t rsw2 ;
  uint8_t rsw3 ;
  pinb = PINB ; // read port
  pinc = PINC ; // for swithces too
  oenc1 = enc1 ;
  oenc2 = enc2 ;
  enc1 = (((pinb >> ENC12_B) &  1) << 1) | ((pinb >> ENC11_B) & 1) ;
  enc2 = (((pinb >> ENC22_B) &  1) << 1) | ((pinb >> ENC21_B) & 1) ;
  enc1_count += enc_to_inc(enc1, oenc1) ;
  enc2_count += enc_to_inc(enc2, oenc2) ;

  // for switches - set low on raw switch low - high if raw switch high and osw low
  // i.e. wait for foreground to ack
  rsw2 = (pinc >> SW2_C) & 1 ; 
  rsw3 = (pinc >> SW3_C) & 1 ;
  if(rsw2 == 0) sw2 = 0 ;
  if((rsw2 == 1) && (osw2 == 0)) sw2 = 1 ;
  rsw3 = rsw3+1 ; // to avoid unused flag
}

// check for update - return increment
// a single "tick" counts 4
int8_t enc_test(int16_t count, int16_t a_count) {
  int16_t delta = count - a_count ;
  if(delta >= 4) return(1) ;
  if(delta <= -4) return(1) ;
  return(0) ;
}

int8_t enc_inc1(void)  {
  if(enc_test(enc1_count, enc1_a_count)) {
    int8_t delta = enc1_count - enc1_a_count ;
    enc1_a_count = enc1_count ;
    return(delta>>2) ;
  }
  return(0) ;
}

int8_t enc_test2(void) {
  return(enc_test(enc2_count, enc2_a_count)) ;
}

// check for update - return increment
int8_t enc_inc2(void)  {
  if(enc_test(enc2_count, enc2_a_count)) {
    int8_t delta = enc2_count - enc2_a_count ;
    enc2_a_count = enc2_count ;
    return(delta>>2) ;
  }
  return(0) ;
}

uint8_t limit_range(int8_t val, uint8_t min, uint8_t max, uint8_t wrap) {
  if(wrap) {
    if(val < min) return(max) ;
    if(val > max) return(min) ;
  }  else {
    if(val < min) return(min) ;
    if(val > max) return(max) ;
  }
  return(val) ;
}

uint16_t enc2_scalar(uint16_t min, uint16_t max, uint16_t inc, uint16_t val) {
  uint16_t v ;
  int8_t dir = enc_inc2() ;
  if((val > min) || (dir >= 0)) v = val + dir * inc ; // don't roll under on min of 0
  else v = val ;
  if(v < min) v = min;
  if(v > max) v = max;
  return(v) ;
}

void lcd_flow_exceeded(void) {
  //lcd_goto(20) ;
  lcd_buff_init() ;
  l_put_string_P(PSTR("Flow Exceeded")) ;
  lcd_buff_row(ROW4) ;
  _delay_ms(1000) ;
  lcd_row4() ;
}

void l_put_mmH2O() {
  l_put_string_P(PSTR(" mm H2O")) ;
}

void l_put_ml() {
  l_put_string_P(PSTR(" mL")) ;
}


// display value currently being entered on LCD row 3
void lcd_row3(void) {
  //  lcd_clear_row(40) ; // row 2
  //lcd_goto(ROW2) ;
  lcd_buff_init() ; // clear buffer
  l_put_string(enc1_vals[enc1_ptr]) ;
  l_put_char(' ') ;
  switch(enc1_ptr) {
  case ENC1_MODE: // Mode
    switch(mode) {
    case MODE_VOL:
      l_put_string_P(PSTR("Volume")) ;
      break ;
    case MODE_VOL_ASSIST:
      l_put_string_P(PSTR("Volume Assist")) ;
      break ;
    case MODE_PRESS:
      l_put_string_P(PSTR("Pressure")) ;
      break ;
    case MODE_PRESS_ASSIST:
      l_put_string_P(PSTR("Pressure Assist")) ;
      break ;
    case MODE_BLEND:
      l_put_string_P(PSTR("Const Flow")) ;
      break ;
    }
    break ;
  case ENC1_TV: // Vol
    l_put_word(vol) ;
    l_put_ml() ;
    break ;
  case ENC1_RR:
    l_put_word_x(bpm,1,10) ;
    l_put_string_P(PSTR(" per min  ")) ;
    break ;
  case ENC1_IE:
    l_put_string_P(PSTR("1:")) ;
    l_put_etoi() ;
    break ;
  case ENC1_MAXP:
    l_put_word(maxp) ;
    l_put_mmH2O() ;
    break ;
  case ENC1_FIO2:
    if(BLEND) {
      l_put_word(fio2) ;
      l_put_char('%') ;
    } else {
      l_put_string_P(PSTR("No Blending")) ;
    }
    break ;
  case ENC1_FLOW:
    if(mode == MODE_BLEND) {
      l_put_word(bflow) ;
    } else {
      l_put_word(flow) ;
    }
    l_put_string_P(PSTR("mL/s")) ;
    break ;
  case ENC1_SPEEP:
    l_put_word(set_peep) ;
    l_put_mmH2O() ;
    break ;
  case ENC1_PCTF:
    l_put_word((100*ps_frac) >> PS_LG_DENOM) ;
    /* old fractional display
    l_put_word(ps_frac) ;
    l_put_char('/') ;
    l_put_word(1<<PS_LG_DENOM) ;*/
    l_put_string_P(PSTR("% Max Flow   ")) ;
    break ;
    /*  case ENC1_SPONT: // spont
    if(p_trig_on == 1) lcd_put_string_P(PSTR("On         ")) ;
    else lcd_put_string_P(PSTR("Off        ")) ;
    break ;*/
  case ENC1_SPONTP: // trigger pressure for spontaneous breathing
    l_put_word(trigp) ;
    l_put_mmH2O() ;
    break ;
  case ENC1_ALARMP: // Alarm PIP
    l_put_word(alarm_pip) ;
    l_put_mmH2O() ;
    break ;
  case ENC1_LMINV: // Low minute vol
    l_put_word(low_minute_vol) ;
    l_put_ml() ;
    break ;
  case ENC1_HMINV: // high minute vol
    l_put_word(high_minute_vol) ;
    l_put_ml() ;
    break ;
  case ENC1_DISCONP: // high minute vol
    l_put_word(pip_disconnect) ;
    l_put_mmH2O() ;
    break ;
  case ENC1_LTV: // last tidal volume
    l_put_word(bvol) ;
    l_put_ml() ;
    break ;
  case ENC1_MVOL: // minute volume
    l_put_word(minute_vol) ;
    l_put_ml() ;
    break ;
  case ENC1_PIP: // peak pressure - last breath
    l_put_word(bpip) ;
    l_put_mmH2O() ;
    break ;
  case ENC1_PEEP: // peak pressure - last breath
    l_put_word(peep) ;
    l_put_mmH2O() ;
    break ;
  case ENC1_INP: // input pressure
    l_put_word(in_p / 10) ;
    l_put_char('.') ;
    l_put_word(in_p % 10)  ;
    l_put_string_P(PSTR("PSI")) ;
    break ;
  case ENC1_O2P: // O2 input pressure
    l_put_word(o2_p / 10) ;
    l_put_char('.') ;
    l_put_word(o2_p % 10)  ;
    l_put_string_P(PSTR("PSI")) ;
    break ;
  case ENC1_INV: // input voltage
    l_put_word(in_v / 10) ;
    l_put_char('.') ;
    l_put_word(in_v % 10)  ;
    l_put_string_P(PSTR("Volts")) ;
    break ;
  case ENC1_VERS:
    l_put_string(VERSION) ;
    break ;
  case ENC1_SAVE: // save and restore - select buffer
  case ENC1_RESTORE: //
    l_put_word(eer_select) ;
    break ;
  case ENC1_ALARM_T: // Test Alarm
    break ;
  case ENC1_CAL: // cal
    l_put_word((uint16_t) ccode) ;
    break ;
  }
  lcd_buff_row(ROW3) ;
}


void encoder_ui(void) {
  int16_t val ; // temporary
  uint16_t offset ; // for save/restore
  uint8_t enc1_range ;

  // on/off switch - no longer used
  /*
  osw1 = sw1 ;
  sw1 = (PINB >> SW1_B) & 1 ;
  if(sw1 == 0) run = 1 ;
  if((osw1 == 0) && (sw1 == 1)) {
    run = 0 ; // this lets the serial start running too.
    set_pwm(0) ;
  }
  */

  // sample pushbuttons
  /*
  osw2 = sw2 ;
  sw2 = (PINC >> SW2_C) & 1 ;
  */

  // leave sw3 old for now
  osw3 = sw3 ;
  sw3 = (PINC >> SW3_C) & 1 ;

  // right pushbutton - reset alarms
  if((sw3 == 0) && (osw3 == 1)) {
    switch(enc1_ptr) {
    case ENC1_RESTORE:
      put_l4(PSTR("Restoring")) ;
      offset = restore_int_to_offset(eer_select) ;
      restore_parameters(offset) ;
      _delay_ms(500) ;
      break;
    case ENC1_SAVE:
      put_l4(PSTR("Saving")) ;
      offset = restore_int_to_offset(eer_select) ;
      save_parameters(offset) ;
      _delay_ms(500) ;
      break;
    case ENC1_ALARM_T:
      alarm_trig(ALARM_TEST) ;
      break ;
    case ENC1_CAL:
      if(ccode == CAL_CCODE_ZERO) { // update zero calibration
	put_l4(PSTR("Discon All Hoses")) ;
	_delay_ms(1000) ;
	nxp_set_zero() ;
	honeywell_set_zero() ;
	siargo_set_zero() ;
	ccode = 0 ;
      } else if (ccode == CAL_CCODE_TAB) { // rebuild table
	put_l4(PSTR("Discon Patient Hoses")) ;
	_delay_ms(1000) ;
	build_f2p_tab(0) ;
	build_f2p_tab(1) ;
	store_f2p_tab(0) ;
	store_f2p_tab(1) ;
	ccode = 0 ;
	enc1_ptr = 0 ;
      }
      break;
    default:
      clear_alarm() ;
      break ;
    }
  }

  // left pushbutton - on/off
  if((sw2 == 0) && (osw2 == 1)) { // toggle run when pushed
    if(run == 0) {
      start_vent() ;
    }
    else {
      stop_vent() ;
    }
    _delay_ms(10) ; // debounce
  }
  osw2 = sw2 ; // acknowledge rtc 

  enc1_ptr = enc1_ptr + enc_inc1() ;

  if(run == 0) enc1_range = NR_ENC1-1 ;
  else enc1_range = RUN_ENC1-1 ;
  
  enc1_ptr = limit_range(enc1_ptr, 0, enc1_range, 0) ; // don't allow save/restore, alarm test, or cal while running
  //lcd_row2() ;
  lcd_row3() ;
  lcd_row4() ;
  if(enc_test2()) {
    switch(enc1_ptr) {
    case ENC1_MODE:
      mode = enc2_scalar(0, MAX_MODES, 1, mode) ;
      break ;
    case ENC1_TV: // Vol
      val = vol ;
      vol = enc2_scalar(MIN_VOL, MAX_VOL, ((vol >= 500) ? INC_VOL2 : INC_VOL), vol) ;
      if(compute_derived() != 1) {
	lcd_flow_exceeded() ;
	vol = val ;
      }
      break ;
    case ENC1_RR: // RR
      val = bpm ;
      bpm = enc2_scalar(MIN_BPM, MAX_BPM, INC_BPM, bpm) ;
      if(compute_derived() != 1) {
	lcd_row3() ;
	lcd_flow_exceeded() ;
	bpm = val ;
      }
      break ;
    case ENC1_IE: // I:E
      val = etoix10 ;
      etoix10 = enc2_scalar(MIN_ETOI, MAX_ETOI, INC_ETOI, etoix10) ;
      if(compute_derived() != 1) {
	lcd_flow_exceeded() ;
	etoix10 = val ;
      }
      break ;
    case ENC1_MAXP: // MaxP
      maxp = enc2_scalar(MIN_MAXP, MAX_MAXP, INC_MAXP, maxp) ;
      break ;
    case ENC1_FIO2: // FIO2
      if(BLEND == 1) {
	if(fio2 == 21) {
	  int8_t dir = enc_inc2() ;
	  if(dir > 0) fio2 = 25 ;
	} else {
	  fio2 = enc2_scalar(MIN_FIO2, MAX_FIO2, INC_FIO2, fio2) ;
	}
      } else {
	fio2 = 21 ;
      }
      break ;
    case ENC1_FLOW:
      if(mode == MODE_BLEND) {
	bflow = enc2_scalar(MIN_BFLOW, MAX_BFLOW,
			    ((bflow < 200) ? INC_BFLOW : INC_BFLOW2), bflow) ;
      }
      break ;
    case ENC1_SPEEP: // set peep
      set_peep = enc2_scalar(MIN_SPEEP, MAX_SPEEP, INC_SPEEP, set_peep) ;
      calculate_spc() ;
      break ;
    case ENC1_PCTF:
      ps_frac = enc2_scalar(MIN_PSFRAC,MAX_PSFRAC,INC_PSFRAC,ps_frac) ; // 1/8 to 7/8 in increments of 1/8
      break ;
      /*case ENC1_SPONT: // Patient initiation - toggle
      p_trig_on = enc2_scalar(0, 1, 1, p_trig_on) ;
      // if(p_trig_on == 0) p_trig_on = 1 ;
      // else p_trig_on = 0 ;
      break ;*/
    case ENC1_SPONTP: // Patient initiation - trigger pressure - negative offset from PEEP
      trigp = enc2_scalar(MIN_TRIGP, MAX_TRIGP, INC_TRIGP, trigp) ;
      break ;
    case ENC1_ALARMP: // Alarm PIP
      alarm_pip = enc2_scalar(MIN_ALARMP, MAX_ALARMP, INC_ALARMP, alarm_pip) ;
      break ;
    case ENC1_LMINV: // Low minute vol
      low_minute_vol = enc2_scalar(MIN_LMV, MAX_LMV, INC_LMV, low_minute_vol) ;
      break ;
    case ENC1_HMINV: // high minute vol
      high_minute_vol = enc2_scalar(MIN_HMV, MAX_HMV, INC_HMV, high_minute_vol) ;
      break ;
    case ENC1_DISCONP:
      pip_disconnect = enc2_scalar(MIN_PIPD, MAX_PIPD, INC_PIPD, pip_disconnect) ;
      break ;
    case ENC1_SAVE:
    case ENC1_RESTORE:
      eer_select = enc2_scalar(0,MAX_EER_OFFSET,1,eer_select) ;
      break ;
    case ENC1_CAL: // Calibrate
      ccode = enc2_scalar(0, 100, 1, ccode) ; // update code

      /*
      lcd_home() ;
      lcd_clear() ;
      lcd_put_string_P(PSTR("Calibrating")) ;
      lcd_goto(ROW2) ;
      lcd_put_string_P(PSTR("Air must be on")) ;
      siargo_set_zero() ;
      nxp_set_zero() ;
      build_f2p_tab() ;
      store_f2p_tab() ;
      enc1_ptr = 0 ;
      */
    default: // for cases where nothing is changed - observe only
      break ; 
    }
    compute_derived() ;
    lcd_row3() ;
  }
}

void ui(void) {

  //set_leds(1) ; // 1 indicates ready
  //  put_string_P(PSTR("PSV Vent:\n\r")) ;
  if(front_panel_lock != 1) {
    lcd_params() ;
    encoder_ui() ; // sample encoders, update display and parms
  }

  if(serial_in_ready()) { // don't wait for keyboard UI
    serial_ui() ;
  }
}


  
  
