// ui_common.c
//
// user interface common routines for open-source proportional solenoid valve ventilator
// (c) Bill Dally, April 2020
//
// Copyright 2020 Bill Dally
// 
// Licensed under the Ventilator Open Source License (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
// ./LICENSE
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <stdio.h>
#include "main.h"
#include "serial.h"
#include "siargo.h"
#include "nxp_pressure.h"
#include "pwm.h"
#include "flow.h"
#include "ui.h"
#include "solenoid.h"
#include "restore.h"

void no_gain(void) { // for open-loop tests
  pgain = 0 ;
  igain = 0 ;
  dgain = 0 ;
}

void set_defaults(void) {
  // user settings
  mode = MODE_VOL ;
  bpm = DEF_BPM ;     // breaths per minute
  vol = DEF_VOL ;   // tidal volume in ml
  etoix10 = DEF_ETOI ; // e:1 ratio times 10
  maxp = DEF_MAXP ;
  fio2 = DEF_FIO2 ;
  set_peep = DEF_SPEEP ;
  alarm_pip = maxp + PIP_ALARM_MARGIN ;
  low_minute_vol = 0 ;
  high_minute_vol = 20000 ;
  pip_disconnect = PIP_DISCONNECT ;
  p_trig_on = 0 ;

  // not settable for now
  trigp = DEF_TRIGP ;

  // gains
  pgain = PGAIN ; // 
  igain = IGAIN ; // 
  dgain = DGAIN ; //
  press_gain = PRESS_GAIN ; //
  press_dgain = PRESS_DGAIN ; //
  press_igain = PRESS_IGAIN ; //

  // rise time
  trise = TRISE ;
  compute_derived() ;

  // pressure
  ps_frac = PS_FRAC ;
}

// compute derived variables
// returns 1 if good, 0 if flow exceeded
uint8_t compute_derived(void)
{
  float ftcy = 60000.0 / (float) bpm ;
  tcy = (uint16_t) ftcy ;
  float fton = (10.0 * (float) tcy / (10.0 + (float) etoix10)) ; // don't subtract trise for now
  fton = fton - INSPIRATORY_PAUSE_TIME ;
  ton = (uint16_t) fton ;
  float ftmax = fton * TMAX_MUL ; // 
  tmax = (uint16_t) ftmax ;
  float fflow = 1000.0 * (float) vol / (fton - (trise/2)) ; // adjust flow to account for rise time
  uint16_t tgap = (tcy-ton) >> 1 ; // half of remaining time
  if(tgap > (ton >> 1)) tgap = ton>>1 ; // half of on time
  //tarm = ton + tgap ;
  tarm = ton ; // no delay to next breath
  if(fflow > MAX_FLOW_MLPS) return(0) ;
  flow = (uint16_t) fflow ;
  flow_mlpm = (uint16_t) (fflow * 60.0) ;
  float fflow_step = fflow * TICK / trise ;
  flow_step = fflow_step ;
  float fvol_limit = ((float) vol) * (1000.0 / TICK) ;
  vol_limit = fvol_limit ;
  ramp_vol = flow * trise/ (2 * TICK) ; // volume for start of ramp
  // expected minute volume
  emvol = vol * bpm ;
  // when to turn on 3-way valve for PEEP
  float fset_peep_open = ((float) set_peep) * PEEP_HIGH_MARGIN ;
  set_peep_open = fset_peep_open ;
  // patient triggering
  if((mode == MODE_VOL_ASSIST) || (mode == MODE_PRESS_ASSIST)) p_trig_on = 1 ;
  else p_trig_on = 0 ;

  // o2 mixing
  // computation is percent, so multiply by 10.24 to get 1024 = 1
  float ffrac_o2_x1k = (10.24 * (((float) fio2 - 21.0)/0.79)) + 0.5 ; 
  frac_o2_x1k = ffrac_o2_x1k ;
  frac_air_x1k = 1024 - frac_o2_x1k ;

  // target air and o2 flow
  int32_t la_flow = ((int32_t) flow * (int32_t)frac_air_x1k) >> 10 ;
  a_flow = la_flow ;
  int32_t lo_flow = ((int32_t) flow * (int32_t)frac_o2_x1k) >> 10 ;
  o_flow = lo_flow ;
  return(1) ;
}

void print_divisor(void) {
  serial_put('/') ; put_word(GAIN_DIVISOR) ; crlf() ;
}

void put_mm_H2O(void) {
  put_string_P(PSTR("mm H2O")) ;
}

void print_values(void)
{
  compute_derived() ;
  crlf() ;
  crlf() ;
  put_string_P(PSTR("Version: ")) ;
  put_string_P(PSTR(VERSION)) ; crlf() ;

  // mode
  switch(mode) {
  case MODE_VOL:
    put_string_P(PSTR("Volume Mode\n\r")) ;
    break ;
  case MODE_VOL_ASSIST:
    put_string_P(PSTR("Volume Assist Mode\n\r")) ;
    break ;
  case MODE_PRESS:
    put_string_P(PSTR("Pressure Mode\n\r")) ;
    break ;
  case MODE_PRESS_ASSIST:
    put_string_P(PSTR("Pressure Assist Mode\n\r")) ;
    break ;
  case MODE_BLEND:
    put_string_P(PSTR("Constant Flow Mode\n\r")) ;
    break ;
  }
  
  put_string_P(PSTR("vol = ")) ; put_word(vol) ; crlf() ;
  put_string_P(PSTR("rr = ")) ; put_byte(bpm) ; crlf() ;
  put_string_P(PSTR("e:i = ")) ; put_byte(etoix10/10) ; dot() ; put_byte(etoix10 % 10) ; crlf() ;
  put_string_P(PSTR("MaxP = ")) ; put_word(maxp) ; put_mm_H2O() ; crlf() ;
  put_string_P(PSTR("FIO2 = ")) ; put_word(fio2) ; serial_put('%') ; crlf() ;
  put_string_P(PSTR("C Flow = ")) ; put_word(bflow) ; put_string_P(PSTR("mL/s")) ; crlf() ;
  put_string_P(PSTR("PEEP = ")) ; put_word(set_peep) ; put_mm_H2O() ; crlf() ;
  put_string_P(PSTR("Percent Flow  = ")) ; put_word(ps_frac) ; put_string_P(PSTR("/8")) ; crlf() ;
  put_string_P(PSTR("Trigger Pressure  = ")) ; put_word(trigp) ; put_mm_H2O() ; crlf() ;
  put_string_P(PSTR("Alarm Pressure  = ")) ; put_word(alarm_pip) ; put_mm_H2O() ; crlf() ;
  put_string_P(PSTR("Low Minute Volume  = ")) ; put_word(low_minute_vol) ; put_string_P(PSTR("mm\n\r")) ;
  put_string_P(PSTR("High Minute Volume  = ")) ; put_word(high_minute_vol) ; put_string_P(PSTR("mm\n\r")) ;
  put_string_P(PSTR("Disconnect Pressure  = ")) ; put_word(pip_disconnect) ; put_mm_H2O() ; crlf() ;
  //  put_string_P(PSTR("Patient initation = ")) ; put_byte(p_trig_on) ; crlf() ;

  put_string_P(PSTR("Insp Hold Time  = ")) ; put_word(thold) ; put_string_P(PSTR("ms\n\r")) ; 
  

  // derived
  put_string_P(PSTR("tcy = ")) ; put_word(tcy) ; put_string_P(PSTR("ms")) ; crlf() ;
  put_string_P(PSTR("ton = ")) ; put_word(ton) ; put_string_P(PSTR("ms")) ; crlf() ;
  put_string_P(PSTR("flow = ")) ; put_word(flow) ; put_string_P(PSTR("mL/sec")) ; crlf() ;
  put_string_P(PSTR("flow = ")) ; put_word(flow_mlpm / 1000) ; dot() ;
  put_word(flow_mlpm % 1000) ; put_string_P(PSTR("L/min")) ; crlf() ;
  put_string_P(PSTR("frac O2, Air = ")) ; put_word(frac_o2_x1k) ; space() ;
  put_word(frac_air_x1k) ; crlf() ;
  put_string_P(PSTR("flow O2, Air = ")) ; put_word(o_flow) ; space() ;
  put_word(a_flow) ; crlf() ;
  
  
  // PID gains
  put_string_P(PSTR("PID = ")) ;
  put_word(pgain) ; space() ;
  put_word(igain) ; space();
  put_word(dgain) ;
  serial_put('/') ; put_word(GAIN_DIVISOR) ; crlf() ;
  // Press gains
  put_string_P(PSTR("Press Gains = ")) ;
  put_word(press_gain) ; space() ;
  put_word(press_igain) ; space() ;
  put_word(press_dgain) ;
  serial_put('/') ; put_word(GAIN_DIVISOR) ; crlf() ;

  //put_string_P(PSTR("vol_limit = ")) ; put_long(vol_limit) ; crlf() ;

  // timing
  put_string_P(PSTR("Tick = ")) ; put_word(TICK) ; crlf() ;

  // sensor cal
  put_string_P(PSTR("Flow,Pres Zero = ")) ; 
  put_word(siargo_flow_zero) ; space() ; put_word(nxp_zero) ; crlf() ;
}

uint8_t in_range(uint16_t val, uint16_t min, uint16_t max, char *msg) {
  if((val >= min)&&(val <= max)) return(1) ;
  else {
    put_word(min) ; put_string_P(PSTR("<=")) ; put_string(msg) ;  put_string_P(PSTR("<=")) ;
    put_word(max) ; crlf() ;
    return(0) ;
  }
}

void flow_exceeded(void) {
  put_string_P(PSTR("Combination of parameters exceeds available flow.\n\r")) ;
}

void one_breath(void) {
  // run one breath with logging
  compute_derived() ;
  put_string_P(PSTR("One Breath")); crlf() ;
  print_values() ;
  time = tcy-10 ; // it will wrap 
  log_ptr = 0 ;
  log_on = 1 ;
  run = 1 ;
  do { // wait while recording happens
    _delay_ms(100) ;
    dot() ;
  } while((log_ptr < (MAX_LOG-1)) || (time < ton));
  crlf() ;
  run = 0 ;
  zero_all_pwm() ;
}

void dump_log(uint16_t len) {
  uint16_t i,j ;
  put_string_P(PSTR("Log, Step = ")); put_byte(log_step) ; crlf() ;
  put_string_P(PSTR("Time, Air_Flow, O2_Flow, Press, CA_Flow, Mvol, A_Err, A_Ierr, Press_IErr, PS_Timer, PState, PS_Flow\n\r") ) ;
  for(i=0;(i<MAX_LOG) && (i<len);i++) {
    put_word(i*log_step*TICK) ; comma() ;
    for(j=0;j<MAX_LOG_CHAN;j++) {
      put_signed(flog[i][j]) ; comma() ;
    }
    crlf() ;
  }
}

// automation of a series of tests
void run_script(void) {
}


void calculate_spc(void) {
  // when to turn off 3-way valve for PEEP
  // do this here to avoid overwriting adjustments
  float fset_peep_close = ((float) set_peep) * PEEP_CLOSE_FRAC ;
  set_peep_close = fset_peep_close ;
}

void start_vent(void) {
  save_settings() ; // save settings to restart on reset - before run = 1
  save_run(1) ; // save that we are running
  run = 1 ;
  gtime = 0 ;
  log_ptr = 0 ;  // log first breath
  log_on = 1 ;
  once = 1 ;
  apnea_count = 0 ;
  if(p_trig_on) {
    time = tarm ; // don't start with mandatory breath in spontaneous mode
    pstate = PS_IDLE ;
  }
  else time = tcy ;
}

void stop_vent(void) {
  run = 0 ;
  save_stop() ; // record that vent has stopped - so it won't restart on reset
  zero_all_pwm() ;
  sol3_off() ;
  // sound alarm for 0.5s in mandatory mode off
  if((sol4_state == 0) && ((mode == MODE_VOL) || (mode == MODE_PRESS))) { 
    sol4_on() ;
    _delay_ms(500) ;
    sol4_off() ;
  }
}

void reset_pf_log(void) {
  pf_log_ptr = 0 ;
  pf_log_cnt = 0 ;
}

void dump_pf_log(void) {
  int16_t i, j, k ;
  // start at begining if log not full
  print_values() ;
  put_string_P(PSTR("Time, Press, Air_Flow, O2_Flow\n\r")) ;
  j = pf_log_cnt ;
  k = 0 ;
  if(pf_log_cnt < MAX_PF_LOG)  i = 0 ;
  else i = pf_log_ptr ;
  do {
    if(i > MAX_PF_LOG) i = 0 ;
    put_word(k * PF_LOG_STEP) ; comma() ;
    put_signed(pf_log[i][0]) ; comma() ; // pressure is signed
    put_word(pf_log[i][1]) ; comma() ;
    put_word(pf_log[i][2]) ; comma() ;
    put_word(pf_log[i][3]) ; crlf() ;
    i += 1 ;
    j -= 1 ;
    k += 1 ;
  } while(j > 0) ;
}

